import json

import numpy as np

from gvec_to_python.base.make_base import make_base
from gvec_to_python.geometry.domain import GVEC_domain
from gvec_to_python.equilibrium.profiles import GVEC_profiles


class GVEC:
    """Primary class of `gvec_to_python` that wraps around everything as a simple interface.
    Loads GVEC data from a .json file, initializes GVEC domain and profiles.

    The class provides the MHD equilibrium calculated by GVEC as callable methods:

    p0, p3              ... equilibrium pressure (0-form, 1-form)
    bv, b1, b2, b_cart  ... equilibrium magentic field (contra-variant, co-variant=1-form, 2-form, Cartesian)
    av, a1, a2, a_cart  ... equilibrium vector potential (contra-variant, co-variant=1-form, 2-form, Cartesian) 

    These methods can be called with float, 1d or 3d array arguments (meshgrid is performed).
    Five different mappings can be chosen:

    1. f_gvec: (s, th, ze) -> (x, y, z)         ... standard gvec mapping
    2. f_pest: (s2, th2, ze2) -> (x, y, z)      ... gvec straight-field-line mapping
    3. f_unit: (s, u, v) -> (x, y, z)           ... gvec with unit cube as logical domain
    4. f_unit_pest : (s, u, v) -> (x, y, z)     ... gvec straight-field-line with unit cube as logical domain
    5. f_zrphi : (s, th, ze) -> (q1, q2, ze)    ... standard gvec to R-Z-phi coordinates

    See GVEC.domian for details. The mapping must be set via the mapping setter method.
    The profiles can be accessed via self.profiles, see GVEC_profiles for details.

    Parameters
    ----------
        json_file : str
            The absolute path to the .json file created by GVEC_Reader.

        mapping : str
            Which mapping to use: "gvec", "pest", "unit", "unit_pest" or "wo_hmap" (see GVEC_domain class).
            Can be changed later using the setter.

        use_nfp : bool
            Whether to use the field period in mappings from the unit cube, i.e. zeta = 2*pi/nfp*v with v in [0, 1] for the toroidal angle.

        use_pyccel : bool
            Whether to use pyccel kernels for matrix operations.
    """

    def __init__(self, json_file: str, mapping='gvec', use_nfp=False, use_pyccel=False):

        # Read GVEC JSON output file.
        with open(json_file) as f:
            data = json.load(f)
        self._data = data

        self._use_pyccel = use_pyccel

        # basis coefficients
        X1_coef = np.array(data['X1']['coef'])
        X2_coef = np.array(data['X2']['coef'])
        LA_coef = np.array(data['LA']['coef'])

        # bases
        X1_base = make_base(data, 'X1')
        X2_base = make_base(data, 'X2')
        LA_base = make_base(data, 'LA')

        # number of field periods (stellarator)
        nfp = data['general']['nfp']

        # Access profile point values (Greville)
        spos_grev = np.array(data['profiles']['greville']['spos'])
        phi_grev = np.array(data['profiles']['greville']['phi'])
        chi_grev = np.array(data['profiles']['greville']['chi'])
        iota_grev = np.array(data['profiles']['greville']['iota'])
        pres_grev = np.array(data['profiles']['greville']['pres'])

        assert len(
            spos_grev) == data['profiles']['greville']['nPoints'],           'Greville points should have the same length as `nPoints`.'
        assert len(
            phi_grev) == data['profiles']['greville']['nPoints'],      'Phi greville profile should have the same length as `nPoints`.'
        assert len(
            chi_grev) == data['profiles']['greville']['nPoints'],      'Chi greville profile should have the same length as `nPoints`.'
        assert len(
            iota_grev) == data['profiles']['greville']['nPoints'],     'Iota greville profile should have the same length as `nPoints`.'
        assert len(
            pres_grev) == data['profiles']['greville']['nPoints'], 'Pressure greville profile should have the same length as `nPoints`.'

        # spline space info for profile interpolation
        sgrid = data["grid"]["sGrid"]
        degree = data['X1']["s_base"]["deg"]

        # domain object
        self._domain = GVEC_domain(X1_base, X2_base, LA_base, X1_coef,
                                   X2_coef, LA_coef, nfp, use_nfp=use_nfp, use_pyccel=use_pyccel)

        # profile object
        self._profiles = GVEC_profiles(
            sgrid, degree, spos_grev, phi_grev, chi_grev, iota_grev, pres_grev)

        # use the mapping setter
        assert mapping in {'gvec', 'pest', 'unit', 'unit_pest', 'wo_hmap'}
        self.mapping = mapping

        # Expose other metadata
        self._minor_radius = data['general']['a_minor']
        self._major_radius = data['general']['r_major']

    @property
    def data(self):
        '''Json data from GVEC run via create_GVEC_json.'''
        return self._data

    @property
    def use_pyccel(self):
        '''Whether to use pyccel kernels for matrix operations.'''
        return self._use_pyccel

    @property
    def domain(self):
        '''Object for all things related to GVEC mappings.'''
        return self._domain

    @property
    def profiles(self):
        '''Object for GVEC radial profiles.'''
        return self._profiles

    @property
    def mapping(self):
        '''Which mapping to use: "gvec", "pest", "unit", "unit_pest" or "wo_hmap" (see GVEC_domain class).'''
        return self._mapping

    @mapping.setter
    def mapping(self, mapping):
        '''Can be used to change the mapping.'''

        # mapping to Cartesian coordinates
        self._f = getattr(self.domain, 'f_' + mapping)
        self._df = getattr(self.domain, 'df_' + mapping)
        self._det_df = getattr(self.domain, 'det_df_' + mapping)
        self._df_inv = getattr(self.domain, 'df_' + mapping + '_inv')
        self._g = getattr(self.domain, 'g_' + mapping)
        self._g_inv = getattr(self.domain, 'g_' + mapping + '_inv')

        # for transformations of a1_gvec and bv_gvec to other logical coordinates than "gvec"
        if mapping in {'gvec', 'wo_hmap'}:
            self.l = None
            self.dl = None
            self.dl_inv = None
        elif mapping == 'pest':
            self.l = self.domain.Pmap
            self.dl = self.domain.dP
            self.dl_inv = self.domain.dP_inv
        elif mapping == 'unit':
            self.l = self.domain.Lmap
            self.dl = self.domain.dL
            self.dl_inv = self.domain.dL_inv
        elif mapping == 'unit_pest':
            self.l = self.domain.PLmap
            self.dl = self.domain.dPL
            self.dl_inv = self.domain.dPL_inv

        self._mapping = mapping

    @property
    def f(self):
        '''The callable map from logical to Cartesian coordiantes, set via the mapping setter.'''
        return self._f

    @property
    def df(self):
        '''The callable Jacobian of the map from logical to Cartesian coordiantes, set via the mapping setter.'''
        return self._df

    @property
    def det_df(self):
        '''The callable Jacobian determinant of the map from logical to Cartesian coordiantes, set via the mapping setter.'''
        return self._det_df

    @property
    def df_inv(self):
        '''The callable inverse Jacobian of the map from logical to Cartesian coordiantes, set via the mapping setter.'''
        return self._df_inv

    @property
    def g(self):
        '''The callable metric tensor of the map from logical to Cartesian coordiantes, set via the mapping setter.'''
        return self._g
    
    @property
    def g_inv(self):
        '''The callable inverse metric tensor of the map from logical to Cartesian coordiantes, set via the mapping setter.'''
        return self._g_inv

    # direct access to MHD equilibrium
    def p0(self, s, a1, a2):
        '''Pressure as 0-form in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A float (single-point evaluation) or 3d numpy array (meshgrid evaluation).
        '''
        return self.profiles.profile(s, a1, a2, name='pressure')

    def p3(self, s, a1, a2):
        '''Pressure as 3-form in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A float (single-point evaluation) or 3d numpy array (meshgrid evaluation).
        '''
        return self.det_df(s, a1, a2) * self.p0(s, a1, a2)

    def p_cart(self, s, a1, a2):
        '''Pressure in Cartesian coordinates, evaluated at one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 2-tuple: first entry is either a float (single-point evaluation) or 3d numpy array (meshgrid evaluation).
                Second entry are the Cartesian coordinates of evaluation points.
        '''
        return self.p0(s, a1, a2), self.f(s, a1, a2)

    def bv(self, s, a1, a2):
        '''Contra-variant components (vector-field) of the magnetic field in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
        '''
        if self.l is None:
            return self._bv_gvec(s, a1, a2)
        else:
            return self.transform(self._bv_gvec(*self.l(s, a1, a2)), s, a1, a2, kind='pull_vector', full_map=False)

    def b1(self, s, a1, a2):
        '''Co-variant components (1-form) of the magnetic field in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
        '''
        return self.transform(self.bv(s, a1, a2), s, a1, a2, kind='v_to_1')

    def b2(self, s, a1, a2):
        '''2-form components of the magnetic field as in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
        '''
        return self.transform(self.bv(s, a1, a2), s, a1, a2, kind='v_to_2')

    def b_cart(self, s, a1, a2):
        '''Cartesian components of the magnetic field evaluated at one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 2-tuple: first entry is a 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
                Second entry are the Cartesian coordinates of evaluation points.
        '''
        return self.transform(self.bv(s, a1, a2), s, a1, a2, kind='push_vector'), self.f(s, a1, a2)

    def av(self, s, a1, a2):
        '''Contra-variant components (vector-field) of the vector potential in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
        '''
        return self.transform(self.a1(s, a1, a2), s, a1, a2, kind='1_to_v')

    def a1(self, s, a1, a2):
        '''Co-variant components (1-form) of the vector potential in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
        '''
        if self.l is None:
            return self._a1_gvec(s, a1, a2)
        else:
            return self.transform(self._a1_gvec(*self.l(s, a1, a2)), s, a1, a2, kind='pull_1form', full_map=False)

    def a2(self, s, a1, a2):
        '''2-form components of the vector potential as in one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
        '''
        return self.transform(self.a1(s, a1, a2), s, a1, a2, kind='1_to_2')

    def a_cart(self, s, a1, a2):
        '''Cartesian components of vector potential evaluated at one of the coordinates "gvec", "pest", "unit" or unit_pest.
        Coordinates can be chosen using the mapping setter.

        Parameters
        ----------
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2. 

        Returns
        -------
            A 2-tuple: first entry is a 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).
                Second entry are the Cartesian coordinates of evaluation points.
        '''
        return self.transform(self.a1(s, a1, a2), s, a1, a2, kind='push_1form'), self.f(s, a1, a2)

    #---------------------------------
    # Internal methods (gvec fields)
    #---------------------------------

    def _a1_gvec(self, s, th, ze):
        '''Vector potential as 1-form (co-variant components) in gvec standard coordinates (map f).

        Parameters
        ----------
            s, th, ze : float or array-like
                Coordinates in [0, 1] x [0, 2*pi]^2. If list or np.array, must be either 1d or 3d (from meshgrid).

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).'''

        phi = self.profiles.profile(s, th, ze, name='phi')
        dphi = self.profiles.profile_derivative(s, th, ze, name='phi')
        chi = self.profiles.profile(s, th, ze, name='chi')

        return (- self.lambda_(s, th, ze) * dphi, phi, -chi)

    def _bv_gvec(self, s, th, ze):
        '''Magnetic field as vector-field (contra-variant components) in gvec standard coordinates (map f).

        Parameters
        ----------
            s, th, ze : float or array-like
                Coordinates in [0, 1] x [0, 2*pi]^2. If list or np.array, must be either 1d or 3d (from meshgrid).

        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).'''

        det_df = self.domain.det_df_gvec(s, th, ze)
        dphi = self.profiles.profile_derivative(s, th, ze, name='phi')
        dchi = self.profiles.profile_derivative(s, th, ze, name='chi')

        return (0.*dphi, (dchi - dphi * self.lambda_ze(s, th, ze)) / det_df, dphi * (1. + self.lambda_th(s, th, ze)) / det_df)

    # lambda function
    def lambda_(self, s, th, ze):
        '''The lambda function from gvec.'''
        return self.domain.LA_base.eval_stz(s, th, ze, self.domain.LA_coef)

    def lambda_s(self, s, th, ze):
        '''s-derivative of the lambda function from gvec.'''
        return self.domain.LA_base.eval_stz_ds(s, th, ze, self.domain.LA_coef)

    def lambda_th(self, s, th, ze):
        '''theta-derivative of the lambda function from gvec.'''
        return self.domain.LA_base.eval_stz_dtheta(s, th, ze, self.domain.LA_coef)

    def lambda_ze(self, s, th, ze):
        '''zeta-derivative of the lambda function from gvec.'''
        return self.domain.LA_base.eval_stz_dzeta(s, th, ze, self.domain.LA_coef)
    
    #---------------
    # Helper methods
    #---------------

    def transform(self, b, s, a1, a2, kind, full_map=True):
        '''Transformations of vector-valued functions under a map m:
        
        pull_vector:    out = dm_inv * b    (dm is the Jacobian of m)
        pull_1form:     out = dm.T * b
        push_vector:    out = dm * b
        push_1form:     out = dm_inv.T * b
        v_to_1:         out = g * b         (g is the metric tensor of m)
        1_to_v:         out = g_inv * b
        v_to_2:         out = det_dm * b
        2_to_v:         out = b / det_dm
        1_to_2:         out = det_dm * g_inv * b
        2_to_1:         out = g * b / det_dm
        
        Parameters
        ----------
            b : tuple[float/array]
                3-tuple of input vector-valued function evaluated as float or array.
                
            s, a1, a2 : float or array-like
                Coordinates in logical space. If list or np.array, must be either 1d or 3d (from meshgrid).
                s in [0, 1] and a1, a2 are angle-like, either in [0, 2*pi]^2 or in [0, 1]^2.
                
            kind : str
                Which transformation to use: "pull_vector", "pull_1form", "pull_2form", "push_vector", "push_1form", or "push_2form".
                
            full_map : bool
                True: use the full map "f" to Cartesian coordinates or False: use a map "l" between logical coordinates.
                Both are set in the mapping setter.
                
        Returns
        -------
            A 3-tuple of either floats (single-point evaluation) or 3d numpy arrays (meshgrid evaluation).'''

        assert isinstance(b, tuple)
        assert len(b) == 3

        # pull to other logical
        if kind == 'pull_vector':
            if full_map:
                raise NotImplementedError
            else:
                if self.mapping == 'gvec':
                    return b
                else:
                    mat = self.dl_inv(s, a1, a2)
                    transpose = False

        elif kind == 'pull_1form':
            if full_map:
                raise NotImplementedError
            else:
                if self.mapping == 'gvec':
                    return b
                else:
                    mat = self.dl(s, a1, a2)
                    transpose = True

        # push to physical
        elif kind == 'push_vector':
            if full_map:
                mat = self.df(s, a1, a2)
                transpose = False
            else:
                raise NotImplementedError

        elif kind == 'push_1form':
            if full_map:
                mat = self.df_inv(s, a1, a2)
                transpose = True
            else:
                raise NotImplementedError

        # transform to other logical
        elif kind == 'v_to_1':
            if full_map:
                mat = self.g(s, a1, a2)
                transpose = False
            else:
                raise NotImplementedError

        elif kind == '1_to_v':
            if full_map:
                mat = self.g_inv(s, a1, a2)
                transpose = False
            else:
                raise NotImplementedError

        elif kind == 'v_to_2':
            if full_map:
                out = self.det_df(s, a1, a2) * np.ascontiguousarray(b)
                return out[0], out[1], out[2]
            else:
                raise NotImplementedError

        elif kind == '2_to_v':
            if full_map:
                out = np.ascontiguousarray(b) / self.det_df(s, a1, a2)
                return out[0], out[1], out[2]
            else:
                raise NotImplementedError

        elif kind == '1_to_2':
            if full_map:
                mat = GVEC_domain.swap_J_axes(GVEC_domain.swap_J_back(self.g_inv(s, a1, a2)) * self.det_df(s, a1, a2))
                transpose = False
            else:
                raise NotImplementedError   

        elif kind == '2_to_1':
            if full_map:
                mat = GVEC_domain.swap_J_axes(GVEC_domain.swap_J_back(self.g(s, a1, a2)) / self.det_df(s, a1, a2))
                transpose = False
            else:
                raise NotImplementedError 

        else:
            raise NotImplementedError      

        # meshgrid evaluation
        if mat.ndim == 5:
            out = GVEC_domain.matvec_5d(mat, np.ascontiguousarray(b), transpose_a=transpose, use_pyccel=self.use_pyccel)
        # single-point evaluation
        else:
            if transpose:
                out = mat.T @ b
            else:
                out = mat @ b

        return out[0], out[1], out[2]
