import pytest

@pytest.mark.parametrize('mapping', ['gvec', 'pest', 'unit', 'unit_pest', 'wo_hmap'])
@pytest.mark.parametrize('use_pyccel', [False, True])   
def test_output_formats(mapping, use_pyccel):
    '''Test whether the return of mapping evaluations has the correct type/dimension.'''

    import os
    import numpy as np
    from gvec_to_python.reader.gvec_reader import create_GVEC_json
    from gvec_to_python import GVEC
    import gvec_to_python as _

    # give absolute paths to the files
    pkg_path = _.__path__[0]
    dat_file_in = os.path.join(
        pkg_path, 'testcases/ellipstell/GVEC_ellipStell_profile_update_State_0000_00010000.dat')
    json_file_out = os.path.join(
        pkg_path, 'testcases/ellipstell/GVEC_ellipStell_profile_update_State_0000_00010000.json')
    create_GVEC_json(dat_file_in, json_file_out, with_spl_coef=True)

    # main object 
    gvec = GVEC(json_file_out, mapping=mapping, use_pyccel=use_pyccel)
    print('gvec.mapping: ', gvec.mapping)

    # test single-point float evaluation
    s = .5
    if 'unit' not in mapping:
        a1 = np.pi
        a2 = np.pi/2
    else:
        a1 = .5
        a2 = .25

    assert isinstance(gvec.f(s, a1, a2)[0], float)
    assert gvec.df(s, a1, a2).shape == (3, 3)
    assert isinstance(gvec.det_df(s, a1, a2), float)
    assert gvec.df_inv(s, a1, a2).shape == (3, 3)
    assert gvec.g(s, a1, a2).shape == (3, 3)
    assert gvec.g_inv(s, a1, a2).shape == (3, 3)

    # test single-point array evaluation
    s = np.array([s])
    a1 = np.array([a1])
    a2 = np.array([a2])

    assert gvec.f(s, a1, a2)[0].shape == (1, 1, 1)
    assert gvec.df(s, a1, a2).shape == (1, 1, 1, 3, 3)
    assert gvec.det_df(s, a1, a2).shape == (1, 1, 1)
    assert gvec.df_inv(s, a1, a2).shape == (1, 1, 1, 3, 3)
    assert gvec.g(s, a1, a2).shape == (1, 1, 1, 3, 3)
    assert gvec.g_inv(s, a1, a2).shape == (1, 1, 1, 3, 3)

    # test 1d array evaluation
    s = np.linspace(.01, 1, 3) # mappings are singular at s=0
    if 'unit' not in mapping:
        a1 = np.linspace(0, 2*np.pi, 4)
        a2 = np.linspace(0, 2*np.pi, 5)
    else:
        a1 = np.linspace(0, 1, 4)
        a2 = np.linspace(0, 1, 5)

    shp3 = (s.size, a1.size, a2.size)
    shp5 = (s.size, a1.size, a2.size, 3, 3)

    assert gvec.f(s, a1, a2)[0].shape == shp3
    assert gvec.df(s, a1, a2).shape == shp5
    assert gvec.det_df(s, a1, a2).shape == shp3
    assert gvec.df_inv(s, a1, a2).shape == shp5
    assert gvec.g(s, a1, a2).shape == shp5
    assert gvec.g_inv(s, a1, a2).shape == shp5

    # test 3d array evaluation
    s, a1, a2 = np.meshgrid(s, a1, a2, indexing='ij')

    assert gvec.f(s, a1, a2)[0].shape == shp3
    assert gvec.df(s, a1, a2).shape == shp5
    assert gvec.det_df(s, a1, a2).shape == shp3
    assert gvec.df_inv(s, a1, a2).shape == shp5
    assert gvec.g(s, a1, a2).shape == shp5
    assert gvec.g_inv(s, a1, a2).shape == shp5

 
def test_values():
    '''Compare values of different mappings provided by GVEC_domain.'''

    import os
    import numpy as np
    from gvec_to_python.reader.gvec_reader import create_GVEC_json
    from gvec_to_python import GVEC
    from gvec_to_python.geometry.domain import GVEC_domain
    import gvec_to_python as _

    # give absolute paths to the files
    pkg_path = _.__path__[0]
    dat_file_in = os.path.join(
        pkg_path, 'testcases/ellipstell/GVEC_ellipStell_profile_update_State_0000_00010000.dat')
    json_file_out = os.path.join(
        pkg_path, 'testcases/ellipstell/GVEC_ellipStell_profile_update_State_0000_00010000.json')
    create_GVEC_json(dat_file_in, json_file_out, with_spl_coef=True)

    # main object 
    gvec = GVEC(json_file_out)

    # compare f and f_pest
    s = np.array([.5])
    th = np.random.rand(10)*2*np.pi
    ze = np.array([np.pi])
    s, th, ze = GVEC_domain.prepare_args(s, th, ze)
    th2 = th + gvec.lambda_(s, th, ze)

    # check Pmap
    assert np.allclose(s, gvec.domain.Pmap(s, th2, ze)[0])
    assert np.allclose(th, gvec.domain.Pmap(s, th2, ze)[1])
    assert np.allclose(ze, gvec.domain.Pmap(s, th2, ze)[2])

    # compare f and f_pest
    assert np.allclose(gvec.domain.f_gvec(s, th, ze)[0], gvec.domain.f_pest(s, th2, ze)[0])
    assert np.allclose(gvec.domain.f_gvec(s, th, ze)[1], gvec.domain.f_pest(s, th2, ze)[1])
    assert np.allclose(gvec.domain.f_gvec(s, th, ze)[2], gvec.domain.f_pest(s, th2, ze)[2])

    # compare f and f_unit
    assert np.allclose(gvec.domain.f_gvec(s, th, ze)[0], gvec.domain.f_unit(s, th/(2*np.pi), ze/(2*np.pi))[0])
    assert np.allclose(gvec.domain.f_gvec(s, th, ze)[1], gvec.domain.f_unit(s, th/(2*np.pi), ze/(2*np.pi))[1])
    assert np.allclose(gvec.domain.f_gvec(s, th, ze)[2], gvec.domain.f_unit(s, th/(2*np.pi), ze/(2*np.pi))[2])

    # compare f_pest and f_unit_pest
    assert np.allclose(gvec.domain.f_pest(s, th2, ze)[0], gvec.domain.f_unit_pest(s, th2/(2*np.pi), ze/(2*np.pi))[0])
    assert np.allclose(gvec.domain.f_pest(s, th2, ze)[1], gvec.domain.f_unit_pest(s, th2/(2*np.pi), ze/(2*np.pi))[1])
    assert np.allclose(gvec.domain.f_pest(s, th2, ze)[2], gvec.domain.f_unit_pest(s, th2/(2*np.pi), ze/(2*np.pi))[2])

 
if __name__ == '__main__':
    test_output_formats('pest', True)
    test_output_formats('unit', True)
    test_values()
