import pytest

@pytest.mark.parametrize('mapping', ['gvec', 'pest', 'unit', 'unit_pest', 'wo_hmap'])
@pytest.mark.parametrize('use_pyccel', [False, True])  
def test_mhd_fields(mapping, use_pyccel):
    '''Test whether the return of mhd field evaluations has the correct type/dimension.'''

    import os
    import numpy as np
    from gvec_to_python.reader.gvec_reader import create_GVEC_json
    from gvec_to_python import GVEC
    import gvec_to_python as _

    # give absolute paths to the files
    pkg_path = _.__path__[0]
    dat_file_in = os.path.join(
        pkg_path, 'testcases/ellipstell/GVEC_ellipStell_profile_update_State_0000_00010000.dat')
    json_file_out = os.path.join(
        pkg_path, 'testcases/ellipstell/GVEC_ellipStell_profile_update_State_0000_00010000.json')
    create_GVEC_json(dat_file_in, json_file_out, with_spl_coef=True)

    # main object 
    gvec = GVEC(json_file_out, mapping=mapping, use_pyccel=use_pyccel)
    print('gvec.mapping: ', gvec.mapping)

    # test single-point float evaluation
    s = .5
    if 'unit' not in mapping:
        a1 = np.pi
        a2 = np.pi/2
    else:
        a1 = .5
        a2 = .25

    assert isinstance(gvec.p0(s, a1, a2), float)
    assert isinstance(gvec.p3(s, a1, a2), float)
    assert np.all([isinstance(gvec.bv(s, a1, a2)[j], float) for j in range(3)])
    assert np.all([isinstance(gvec.b1(s, a1, a2)[j], float) for j in range(3)])
    assert np.all([isinstance(gvec.b2(s, a1, a2)[j], float) for j in range(3)])
    assert np.all([isinstance(gvec.b_cart(s, a1, a2)[i][j], float) for j in range(3) for i in range(2)])
    assert np.all([isinstance(gvec.av(s, a1, a2)[j], float) for j in range(3)])
    assert np.all([isinstance(gvec.a1(s, a1, a2)[j], float) for j in range(3)])
    assert np.all([isinstance(gvec.a2(s, a1, a2)[j], float) for j in range(3)])
    assert np.all([isinstance(gvec.a_cart(s, a1, a2)[i][j], float) for j in range(3) for i in range(2)])

    # test single-point array evaluation
    s = np.array([s])
    a1 = np.array([a1])
    a2 = np.array([a2])

    assert gvec.p0(s, a1, a2).shape == (1, 1, 1)
    assert gvec.p3(s, a1, a2).shape == (1, 1, 1)
    assert np.all([gvec.bv(s, a1, a2)[j].shape == (1, 1, 1) for j in range(3)])
    assert np.all([gvec.b1(s, a1, a2)[j].shape == (1, 1, 1) for j in range(3)])
    assert np.all([gvec.b2(s, a1, a2)[j].shape == (1, 1, 1) for j in range(3)])
    assert np.all([gvec.b_cart(s, a1, a2)[i][j].shape == (1, 1, 1) for j in range(3) for i in range(2)])
    assert np.all([gvec.av(s, a1, a2)[j].shape == (1, 1, 1) for j in range(3)])
    assert np.all([gvec.a1(s, a1, a2)[j].shape == (1, 1, 1) for j in range(3)])
    assert np.all([gvec.a2(s, a1, a2)[j].shape == (1, 1, 1) for j in range(3)])
    assert np.all([gvec.a_cart(s, a1, a2)[i][j].shape == (1, 1, 1) for j in range(3) for i in range(2)])

    # test 1d array evaluation
    s = np.linspace(.01, 1, 4) # mappings are singular at s=0
    if 'unit' not in mapping:
        a1 = np.linspace(0, 2*np.pi, 5)
        a2 = np.linspace(0, 2*np.pi, 6)
    else:
        a1 = np.linspace(0, 1, 4)
        a2 = np.linspace(0, 1, 5)

    shp3 = (s.size, a1.size, a2.size)

    assert gvec.p0(s, a1, a2).shape == shp3
    assert gvec.p3(s, a1, a2).shape == shp3
    tmp = gvec.bv(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.b1(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.b2(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.b_cart(s, a1, a2)
    assert np.all([tmp[i][j].shape == shp3 for j in range(3) for i in range(2)])
    tmp = gvec.av(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.a1(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.a2(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.a_cart(s, a1, a2)
    assert np.all([tmp[i][j].shape == shp3 for j in range(3) for i in range(2)])

    # test 3d array evaluation
    s, a1, a2 = np.meshgrid(s, a1, a2, indexing='ij')

    assert gvec.p0(s, a1, a2).shape == shp3
    assert gvec.p3(s, a1, a2).shape == shp3
    tmp = gvec.bv(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.b1(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.b2(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.b_cart(s, a1, a2)
    assert np.all([tmp[i][j].shape == shp3 for j in range(3) for i in range(2)])
    tmp = gvec.av(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.a1(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.a2(s, a1, a2)
    assert np.all([tmp[j].shape == shp3 for j in range(3)])
    tmp = gvec.a_cart(s, a1, a2)
    assert np.all([tmp[i][j].shape == shp3 for j in range(3) for i in range(2)])


if __name__ == '__main__':
    test_mhd_fields('gvec', use_pyccel=False)
    test_mhd_fields('pest', use_pyccel=True)